@@ -1,4 +1,3 @@
-# -*- coding: utf-8 -*-
 
 # Copyright 2012 Jaap Karssenberg <jaap.karssenberg@gmail.com>
 
@@ -17,27 +16,33 @@ Note that this can not replace "zim --export" because no effort is
 done here to resolve links. Main purpose is testing.
 '''
 
+import os
 import sys
 import logging
 
+from zim.newfs import localFileOrFolder
 from zim.formats import *
 
 
 if __name__ == '__main__':
-	if len(sys.argv) not in (2, 3):
-			print 'Usage: python -m zim.formats format [format]'
-			print '\tWill read from stdin and output to stdout'
+	if len(sys.argv) not in (2, 3, 4):
+			print('Usage: python -m zim.formats format [format] [source_dir]')
+			print('\tWill read from stdin and output to stdout')
 			sys.exit(1)
 
 
 	logging.basicConfig()
 
 	inputformat = sys.argv[1]
-	if len(sys.argv) == 3:
+	if len(sys.argv) == 4:
 		outputformat = sys.argv[2]
+		source_dir = localFileOrFolder(sys.argv[3], pwd=os.getcwd())
+	elif len(sys.argv) == 3:
+		outputformat = sys.argv[2]
+		source_dir = None
 	else:
 		outputformat = '__XML__'
-
+		source_dir = None
 
 	input = sys.stdin.read()
 
@@ -47,6 +52,7 @@ if __name__ == '__main__':
 	if outputformat == '__XML__':
 		sys.stdout.write(tree.tostring())
 	else:
-		dumper = get_dumper(outputformat, linker=StubLinker())
+		linker = StubLinker(source_dir=source_dir)
+		dumper = get_dumper(outputformat, linker=linker)
 		lines = dumper.dump(tree)
-		sys.stdout.write(''.join(lines).encode('utf-8'))
+		sys.stdout.write(''.join(lines))
